#ifndef GLOBAL_BUFFER_H
#define GLOBAL_BUFFER_H

#include <systemc.h>
#include <tlm.h>
#include <vector>
#include <iostream>
#include <fstream>
#include <sstream>
#include <tlm_utils/simple_initiator_socket.h>
#include <tlm_utils/simple_target_socket.h>
#include "utils/configuration.h"
#include "utils/noc_logger.h"
#include <cstdlib>  // For rand()
#include <ctime>    // For time()
#include <string>
#include <queue>
#include "utils/memory_manager.h"

using namespace tlm;
using namespace sc_core;

class GlobalBuffer : public sc_module {
public:
    // Constructor
    SC_HAS_PROCESS(GlobalBuffer);
    
// TLM sockets for communication with interposer interfaces   
    tlm_utils::simple_initiator_socket<GlobalBuffer> SingleChiplet_from_GB_sockets[CHIPLET_NO];   // Sending to chiplets
    tlm_utils::simple_initiator_socket<GlobalBuffer> CrossChiplet_from_GB_sockets[CHIPLET_NO];   // Sending to chiplets
    tlm_utils::simple_target_socket<GlobalBuffer> receive_to_GB_sockets[CHIPLET_NO];  // Receiving Packets from chiplets
//Sockets connecting GB to NI of GB   
    tlm_utils::simple_target_socket<GlobalBuffer> ni_to_gb_receive_socket; //changed naming
    tlm_utils::simple_initiator_socket<GlobalBuffer> gb_to_ni_send_socket;
//FiFos
    sc_core::sc_fifo<PESendPayload*> SingleChiplet_fifo;  // FIFO for incoming SingleChiplet transactions from ni
    sc_core::sc_fifo<PESendPayload*> CrossChiplet_fifo;  // FIFO for CrossChiplet transactions - incoming weight kernels from ni and packets from PE
    sc_core::sc_fifo<PESendPayload*> OF_fifo;  // FIFO for OF transactions from the pe

    MemoryManager mm;

    explicit GlobalBuffer(sc_module_name name, int gb_id, int num_chiplets, int num_pes);

private:

    // Data members
    int num_chiplets;     // Number of chiplets (interposer interfaces)
    int num_pes;          // Number of PEs in total
    int gb_id;

    // Function to handle received TLM  from interposer interface
    void receive_to_gb_b_transport(tlm::tlm_generic_payload& trans, sc_time& delay);

    //non-blocking communication with NI of GB
    tlm::tlm_sync_enum gb_nb_transport_fw(tlm::tlm_generic_payload& trans, tlm::tlm_phase& phase, sc_time& delay);
    
    // TLM Communication methods
    void single_chiplet_communication(); //To send single chiplet data to interposer interface

    void cross_chiplet_communication(); // To send cross chiplet data to interposer interface

    void OF_communication(); // To process the received OF from PEs

    void log_info(std::string msg);

    void log_error(std::string msg);

};

#endif // GLOBAL_BUFFER_H