#include "O_E_converter.h"
#include <iostream>

// Constructor implementation
O_E_converter::O_E_converter(sc_module_name name)
    : sc_module(name) {
    sc_report_handler::set_actions(OEC_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
    // Register b_transport methods for GB and Processing Elements
    ci_SingleChiplet_receive_socket.register_b_transport(this, &O_E_converter::ci_SingleChiplet_b_transport);
    ci_CrossChiplet_receive_socket.register_b_transport(this, &O_E_converter::ci_CrossChiplet_b_transport);
    //Register b_transport for incoming packets from ni
    ni_to_oec_receive_socket.register_b_transport(this, &O_E_converter::ni_b_transport);
    //SC_THREAD(chiplet_send_transaction);
}

// Handle incoming transactions from the chiplet interface
void O_E_converter::ci_SingleChiplet_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        wait(IF_DELAY);
        log_error(std::to_string(O_E_converter_id) + " Invalid transaction type (not PESendPayload)." );
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    PESendPayload* oec_trans1 = new PESendPayload();

    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    sc_time sent_time = orig_trans->timestamp;
    oec_trans1->data = orig_trans->data;
    oec_trans1->src_pe = orig_trans->src_pe;
    oec_trans1->src_chiplet = orig_trans->src_chiplet;
    oec_trans1->dest_pe = orig_trans->dest_pe;
    oec_trans1->dest_chiplet = orig_trans->dest_chiplet;
    oec_trans1->type_flag = orig_trans->type_flag;
    oec_trans1->update_timestamp(current_time);
    oec_trans1->set_address(orig_trans->get_address());
    oec_trans1->set_command(orig_trans->get_command());
    oec_trans1->set_data_ptr(reinterpret_cast<unsigned char*>(&oec_trans1->data));
    oec_trans1->set_data_length(orig_trans->get_data_length());
    oec_trans1->set_streaming_width(orig_trans->get_streaming_width());
    oec_trans1->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    oec_trans1->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = oec_trans1->get_address(); // address is used to determine wavelength
    // Single-chiplet communication (λM)
    if (address >= 0x100 && address < 0x200) {
        // Single-chiplet communication
        int chiplet_index = address - 0x100;
        int chiplet_id = chiplet_index - PE_PER_CHIPLET;
        int single_wavelength = chiplet_index;

        log_info(std::to_string(O_E_converter_id) + " Single-chiplet communication, wavelength: " + std::to_string(single_wavelength));
        
        trans.set_response_status(TLM_OK_RESPONSE);  // Set response status
        
        log_info(std::to_string(O_E_converter_id) + " Received input feature with id: " 
                + std::to_string(oec_trans1->data->id) + "as single chiplet transaction, with Wavelength: " 
                + std::to_string(single_wavelength));
        sc_time oec_delay = IF_DELAY * O_E_converter_id;
        wait(oec_delay); //A 0.5 ns delay between PEs or delay from chiplet interface to PEs is modeled here
        //wait(IF_DELAY); //The simple delay for input features when we neglect the delays on chiplets between PEs
        //double od = oec_delay.to_double();
        //FUNCTION_INFO("Delay at OEC for Single chiplet", (od) );
        // Forward the transaction to the Processing Element 
        sc_core::sc_time local_delay_oec_single = SC_ZERO_TIME;
        ni_SingleChiplet_send_transaction(*oec_trans1, local_delay_oec_single, single_wavelength);

    } else {
        log_error(std::to_string(O_E_converter_id) + " Invalid or unrecognized address: " 
                    + std::to_string(address));
    }

    //delete oec_trans1;  // Clean up dynamically allocated memory    
}

void O_E_converter::ni_SingleChiplet_send_transaction(PESendPayload& trans, sc_time& delay, int single_wavelength){
    log_info(std::to_string(O_E_converter_id) + " Forwarding Single Chiplet transaction with id: " 
            + std::to_string(trans.data->id) + " with Wavelength: " + std::to_string(single_wavelength) 
            + " to Processing Element: " + std::to_string(O_E_converter_id));
    ni_SingleChiplet_send_socket->b_transport(trans, delay);  // Send transaction to Processing Element    
}

// Handle incoming Cross chiplet transaction from the Global Buffer
void O_E_converter::ci_CrossChiplet_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        wait(WK_DELAY);
        log_error(std::to_string(O_E_converter_id) + " Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    PESendPayload* oec_trans2 = new PESendPayload();
    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    sc_time sent_time = orig_trans->timestamp; // Log the time when the core sent the transaction
    oec_trans2->data = orig_trans->data;
    oec_trans2->src_pe = orig_trans->src_pe;
    oec_trans2->src_chiplet = orig_trans->src_chiplet;
    oec_trans2->dest_pe = orig_trans->dest_pe;
    oec_trans2->dest_chiplet = orig_trans->dest_chiplet;
    oec_trans2->type_flag = orig_trans->type_flag;
    oec_trans2->update_timestamp(current_time);
    oec_trans2->set_address(orig_trans->get_address());
    oec_trans2->set_command(orig_trans->get_command());
    oec_trans2->set_data_ptr(reinterpret_cast<unsigned char*>(&oec_trans2->data));
    oec_trans2->set_data_length(orig_trans->get_data_length());
    oec_trans2->set_streaming_width(orig_trans->get_streaming_width());
    oec_trans2->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    oec_trans2->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = oec_trans2->get_address(); // address is used to determine wavelength
    // Cross-chiplet communication (λ0, λ1, λ2, ...)
    if (address >= 0x200) {
        int cross_wavelength = address - 0x200;
        log_info("Cross-chiplet communication, wavelength : " + std::to_string(cross_wavelength));
        wait(WK_DELAY);
        
        log_info(std::to_string(O_E_converter_id) + " Received cross-chiplet data: " 
                + std::to_string(oec_trans2->data->id) + " with Wavelength: " 
                + std::to_string(cross_wavelength));
        trans.set_response_status(TLM_OK_RESPONSE);    
        // Forward modified data to Processing Element
        sc_time local_delay_oec_cross = SC_ZERO_TIME;
        chiplet_CrossChiplet_send_transaction(*oec_trans2, local_delay_oec_cross, cross_wavelength);
    } else {
        log_error(std::to_string(O_E_converter_id) + " Invalid or unrecognized address: " 
                    + std::to_string(address));    
    }

    //delete oec_trans2;  // Clean up dynamically allocated memory   
}

void O_E_converter::chiplet_CrossChiplet_send_transaction(PESendPayload& trans, sc_time& delay, int cross_wavelength) {
    //std::cout << "check: The delay before transmitting cross chiplet data to PE: " << delay << std::endl;
    ni_CrossChiplet_send_socket->b_transport(trans, delay);  // Send transaction to Processing Element
    log_info(std::to_string(O_E_converter_id) + " Forwarding Cross chiplet transaction: " 
            + std::to_string(trans.data->id) + " with Wavelength: " + std::to_string(cross_wavelength) 
            + " to Processing Element: " + std::to_string(cross_wavelength));  
}

//b_transport function to receive from the NI
void O_E_converter::ni_b_transport(tlm_generic_payload& trans, sc_time& delay) {
    PESendPayload* orig_trans = dynamic_cast<PESendPayload*>(&trans);
    if (!orig_trans) {
        wait(OF_DELAY);
        log_error(std::to_string(O_E_converter_id) + "  Invalid transaction type (not PESendPayload).");
        trans.set_response_status(TLM_GENERIC_ERROR_RESPONSE);
        return;
    }

    PESendPayload* rx_trans1 = new PESendPayload();

    sc_time current_time = sc_time_stamp();
    //manually copy the transaction payload for future use
    sc_time sent_time = orig_trans->timestamp;
    rx_trans1->data = orig_trans->data;
    rx_trans1->src_chiplet = orig_trans->src_chiplet;
    rx_trans1->src_pe = orig_trans->src_pe;
    rx_trans1->src_pe = orig_trans->src_pe;
    rx_trans1->src_chiplet = orig_trans->src_chiplet;
    rx_trans1->dest_pe = orig_trans->dest_pe;
    rx_trans1->dest_chiplet = orig_trans->dest_chiplet;
    rx_trans1->type_flag = orig_trans->type_flag;
    rx_trans1->update_timestamp(current_time);
    rx_trans1->set_address(orig_trans->get_address());
    rx_trans1->set_command(orig_trans->get_command());
    rx_trans1->set_data_ptr(reinterpret_cast<unsigned char*>(&rx_trans1->data));
    rx_trans1->set_data_length(orig_trans->get_data_length());
    rx_trans1->set_streaming_width(orig_trans->get_streaming_width());
    rx_trans1->set_byte_enable_ptr(orig_trans->get_byte_enable_ptr());
    rx_trans1->set_dmi_allowed(orig_trans->is_dmi_allowed());

    int address = rx_trans1->get_address(); // address is used to determine wavelength

    int chiplet_index = address - 0x300;
   
    int unicast_wavelength = chiplet_index;
    wait(OF_DELAY);
    log_info(std::to_string(O_E_converter_id) + " Optical to Electrical converter received Packet from NI: " 
            + std::to_string(rx_trans1->src_pe) + " with data id: " + std::to_string(rx_trans1->data->id));
    trans.set_response_status(TLM_OK_RESPONSE);  // Set response status

    // Forward the transaction to the Chiplet interface 
    sc_time local_delay_oec_OF = SC_ZERO_TIME;
    chiplet_send_transaction(*rx_trans1, local_delay_oec_OF);
    //delete rx_trans1;  // Clean up dynamically allocated memory
    
}

void O_E_converter::chiplet_send_transaction(PESendPayload& trans, sc_time& delay){
    //std::cout << "check: The delay before transmitting OF to CI: " << delay << std::endl;
    ci_send_socket->b_transport(trans, delay);    
    log_info(std::to_string(O_E_converter_id) + " Forwarding the Packets from NI: " + std::to_string(trans.src_pe) 
            + " to chiplet interface: " + std::to_string(trans.src_chiplet));
}

void O_E_converter::log_info(std::string msg){
    SC_REPORT_INFO(OEC_LOG, msg.c_str());
}

void O_E_converter::log_error(std::string msg){
    SC_REPORT_ERROR(OEC_LOG, msg.c_str());
}