/*******************************************************************************
 * Copyright (C) 2024 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/

// NocTlm.h
// #ifndef NOCTLM_H
// #define NOCTLM_H
#pragma once

#ifndef SC_INCLUDE_DYNAMIC_PROCESSES
#define SC_INCLUDE_DYNAMIC_PROCESSES
#endif

#include <systemc.h>
#include <tlm.h>
#include <tlm_utils/simple_initiator_socket.h>
#include <tlm_utils/simple_target_socket.h>
#include <queue>
#include <iostream>
#include <vector>
#include <string>
#include <map>


#include "ratatoskrUtils/utils/Structures.h"
#include "ratatoskrUtils/utils/GlobalReport.h"
#include "ratatoskrUtils/utils/GlobalResources.h"

#include "networkInterface/NetworkInterfacePE.h"
#include "networkInterface/NetworkInterfaceGB.h"

#include "GB/Global_Buffer.h"
#include "Interposer/Interposer_Interface.h"
#include "Chiplet/Chiplet_Interface.h"
#include "Optical_Electrical/O_E_converter.h"

#include "utils/configuration.h" 
#include "utils/token_manager.h"
#include "utils/noc_logger.h"
#include "utils/utils.h"

using namespace tlm;
using namespace sc_core;

class NocTlm : public sc_core::sc_module {
public:
  
    SC_HAS_PROCESS(NocTlm);

    // Constructor and destructor
    explicit NocTlm(sc_module_name, std::string);
    ~NocTlm() override;

    uint8_t max_pos[3];
    std::map<int, int> mapNodePERouter;

    // Define the number of chiplets and processing elements
    const int num_chiplets = CHIPLET_NO;
    const int num_pes = PE_PER_CHIPLET;
    const int total_pes = PE_NO;

    void initializeGlobalResources(std::string config_folder);

    GlobalResources& globalResources = GlobalResources::getInstance();

    // Create arrays of Interposer Interface instances, and other modules
    InterposerInterface* interposers[CHIPLET_NO];
    ChipletInterface* chiplets[CHIPLET_NO];
    O_E_converter* oec[CHIPLET_NO][PE_PER_CHIPLET];
    TokenManager* tokens[CHIPLET_NO];
    GlobalBuffer* gb;
    NetworkInterfacePE* nis[CHIPLET_NO][PE_PER_CHIPLET];
    NetworkInterfaceGB* ni_gb;


    // Method to initialize the routers and nis
    void setup_modules();

    void log_info(std::string msg);

    void log_warn(std::string msg);

    void log_error(std::string msg);

};

// #endif // NOCTLM_H
